/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>

#include "cntree.h"

cntree *
cntree_new(void)
{
  cntree *ntree;
  
  ntree = (cntree *) calloc(1, sizeof(cntree));

  return ntree;
}

cntree *
cntree_new_with_data(void *data)
{
  cntree *ntree = cntree_new();
  if (ntree == NULL)
    return NULL;
  ntree->data = data;
  
  return ntree;
}

/* Recursively traverse the tree.
 * If non NULL, free_callback() is meant to free the data part of the tree.
 * The tree structures are freeed.
 */
void
cntree_free(cntree *ntree, void (*free_callback)(cntree *atree, void *adata),
	    void *data)
{
  clistiter *iter;
  cntree *atree;
  
  if (ntree == NULL)
    return;
  if (ntree->children)
    {
      for (iter = clist_begin(ntree->children); iter; iter = clist_next(iter))
	{
	  atree = (cntree *) clist_content(iter);
	  cntree_free(atree, free_callback, data);
	}
      clist_free(ntree->children);
    }
  if (free_callback)
    free_callback(ntree, data);
  free(ntree);
}

clist *
cntree_get_children(cntree *ntree)
{
  return ntree->children;
}

void
cntree_set_children(cntree *parent, clist *children)
{
  clistiter *iter;
  cntree *child;
  parent->children = children;
  if (children)
    {
      for (iter = clist_begin(children); iter; iter = clist_next(iter))
	{
	  child = (cntree *) clist_content(iter);
	  child->parent = parent;
	}  
    }
}

int
cntree_append_child(cntree *parent, cntree *child)
{
  if (parent->children == NULL)
    parent->children = clist_new();
  if (parent->children && clist_append(parent->children, child) == 0)
    {
      child->parent = parent;
      return 0;
    }
  return -1;
}

int
cntree_prepend_child(cntree *parent, cntree *child)
{
  if (clist_prepend(parent->children, child) == 0)
    {
      child->parent = parent;
      return 0;
    }
  return -1;
}

/*
 * Find child amongs parent's children and remove it.
 * Return 0 on success and -1 if the child was not found.
 */
int
cntree_remove_child(cntree *parent, cntree *child)
{
  clistiter *iter;

  if (parent->children)
    {
      for (iter = clist_begin(parent->children); iter; iter = clist_next(iter))
	{
	  if (clist_content(iter) == (void *) child) {
	    clist_delete(parent->children, iter);
	    child->parent = NULL;	
	    return 0;
	  }
	}
    }
  return -1;
}

void
cntree_set_data(cntree *ntree, void *data)
{
  ntree->data = data;
}

void *
cntree_get_data(cntree *ntree)
{
  return ntree->data;
}

int
cntree_get_children_count(cntree *ntree)
{
  return (ntree->children ? clist_count(ntree->children) : 0);
}

/*
 * For every item in the tree and its descendents, call a callback
 * function with the current tree and a user defined data as arguments.
 * The function must return zero to continue traversal.
 * Traversal is not done level by level but first child's children are
 * traversed first.
 */
int
cntree_for_each(cntree *ntree, int (*callback)(cntree *atree, void *adata),
		void *data)
{
  clistiter *iter;
  cntree *atree;
  int ret;
  
  ret = callback(ntree, data);
  if (ntree->children)
    {
      for (iter = clist_begin(ntree->children); iter && !ret;
	   iter = clist_next(iter))
	{
	  atree = (cntree *) clist_content(iter);
	  ret = cntree_for_each(atree, callback, data);
	}
    }
  return ret;
}

int
cntree_for_each_child(cntree *ntree, 
		      int (*callback)(cntree *atree, void *adata),
		      void *data)
{
  clistiter *iter;
  cntree *atree;
  int ret = 0;

  if (ntree->children)
    {
      for (iter = clist_begin(ntree->children); iter && !ret;
	   iter = clist_next(iter))
	{
	  atree = (cntree *) clist_content(iter);
	  ret = callback(atree, data);
	}
    }
  return ret;
}

cntree *
cntree_get_root(cntree *ntree)
{
  if (ntree->parent == NULL)
    return ntree;
  else
    return cntree_get_root(ntree->parent);
}

cntree *
cntree_is_child(cntree *parent, void *data)
{
  clistiter *iter;
  cntree *child;
  
  if (parent->children)
    {
      for (iter = clist_begin(parent->children); iter; iter = clist_next(iter))
	{
	  child = (cntree *) clist_content(iter);
	  if (child && child->data == data)
	    return child;
	}
    }
  return NULL;
}
