/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002, 2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>

#include <libetpan/libetpan.h>

#include "logging.h"
#include "x_alloc.h"
#include "message.h"
#include "folder.h"

#include "vmessage.h"

pthread_mutex_t vmessage_mutex = PTHREAD_MUTEX_INITIALIZER;

int
VMessage_mailmessage_init(mailmessage *msg, 
			  struct mail_flags *msg_flags,
			  struct mailimf_single_fields *msg_fields,
			  size_t msg_size);

static mailmessage *
VMessage_mailmessage_copy(mailmessage *msg);

static int
VMessage_clone_orig_date(struct mailimf_orig_date *msg_orig_date, 
			 struct mailimf_orig_date **orig_date);

static int
VMessage_clone_from(struct mailimf_from *msg_from,
		    struct mailimf_from **from);

static int
VMessage_clone_sender(struct mailimf_sender *msg_sender,
		      struct mailimf_sender **sender);

static int
VMessage_clone_reply_to(struct mailimf_reply_to *msg_reply_to,
			struct mailimf_reply_to **reply_to);

static int
VMessage_clone_to(struct mailimf_to *msg_to,
		  struct mailimf_to **to);

static int
VMessage_clone_cc(struct mailimf_cc *msg_cc,
		  struct mailimf_cc **cc);

static int
VMessage_clone_bcc(struct mailimf_bcc *msg_bcc,
		   struct mailimf_bcc **bcc);

static int
VMessage_clone_message_id(struct mailimf_message_id *msg_message_id,
			  struct mailimf_message_id **message_id);

static int
VMessage_clone_in_reply_to(struct mailimf_in_reply_to *msg_in_reply_to,
			   struct mailimf_in_reply_to **in_reply_to);

static int
VMessage_clone_references(struct mailimf_references *msg_references,
			  struct mailimf_references **references);

/* Function definitions */
XePVMessage *
a_VMessage_new_from_message(XePMessage *message)
{
  XePVMessage *vmessage;
  char *vkey;
  size_t folder_id_len;
  size_t message_uid_len;
  chashdatum key;
  chashdatum data;
  mailmessage *msg;
  XePFolder *folder;

  vmessage = x_alloc(1, XePVMessage);

  if (vmessage == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }

  vmessage->folder = message->folder;
  /* Key is message->uid */

  vmessage->msg = mailmessage_new();
  if (vmessage->msg == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      goto free_vmessage;
    }

  msg = message->msg;
  if (VMessage_mailmessage_init(vmessage->msg, 
				msg->msg_flags, 
				&msg->msg_single_fields, 
				msg->msg_size))
    {
      goto free_vmsg;
    }

  key.data = message->uid;
  key.len = strlen(message->uid);
  data.data = (char *) vmessage;
  data.len = sizeof(vmessage);
  vmessage->vfolders = clist_new();
  if (vmessage->vfolders == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("clist_new() failed: %s", strerror(errno));
      goto free_vmsg;
    }

  folder = message->folder;

  if (chash_set(folder->vmessages, &key, &data, NULL))
    {
      A_LOGGING_SYSTEM_ERROR("chash_set() failed for vmessage %s@%s",
			     message->uid, message->folder->id);
      goto free_vfolders;
    }

  return vmessage;

 free_vfolders:
  clist_free(vmessage->vfolders);
 free_vmsg:
  mailmessage_free(vmessage->msg);
 free_vmessage:
  free(vmessage);
      return NULL;
}

void
a_VMessage_free(XePVMessage *vmessage)
{
  clistiter *iter;
  for (iter = clist_begin(vmessage->vfolders); iter; iter = iter->next)
    {
      /* XXX: vfolders must not retain pointers to mailmessage pointing
       * to the vmessage */
      x_free(iter->data);
    }
  clist_free(vmessage->vfolders);

  mailmessage_free(vmessage->msg);

  free(vmessage);
}

int
VMessage_mailmessage_init(mailmessage *msg, 
			  struct mail_flags *msg_flags,
			  struct mailimf_single_fields *msg_fields,
			  size_t msg_size)
{
  struct mail_flags *flags;
  clistiter *iter;

  /* Copy size */
  msg->msg_size = msg_size;

  /* Copy flags */
  if (msg_flags)
    {
      clist *extension;
      int ok = -1;

      extension = clist_new();
      if (flags == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("clist_new(): %s", strerror(errno));
	  return -1;
	}

      for (iter = clist_begin(msg_flags->fl_extension); iter && ok; iter = iter->next)
	{
	  char *ext_flag;
	  ext_flag = strdup((char *)iter->data);
	  if (ext_flag == NULL || clist_append(extension, ext_flag))
	    ok = 0;
	}

      if (!ok || (flags = mail_flags_new(msg_flags->fl_flags, extension)) == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));

	  for (iter = clist_begin(extension); iter; iter = iter->next)
	    {
	      free(iter->data);
	    }
	  free(extension);
	  return -1;
	}
    }
  else
    {
      flags = NULL;
    }
  msg->msg_flags = flags;
      
  /* Copy single fields */
  if (msg_fields)
    {
      /* orig_date */
      struct mailimf_orig_date *orig_date;
      struct mailimf_from *from;
      struct mailimf_sender *sender;
      struct mailimf_reply_to *reply_to;
      struct mailimf_to *to;
      struct mailimf_cc *cc;
      struct mailimf_bcc *bcc;
      struct mailimf_message_id *message_id;
      struct mailimf_in_reply_to *in_reply_to;
      struct mailimf_references *references;
      int r;

      r = VMessage_clone_orig_date(msg_fields->fld_orig_date, &orig_date);
      msg->msg_single_fields.fld_orig_date = orig_date;
      
      r = VMessage_clone_from(msg_fields->fld_from, &from);
      msg->msg_single_fields.fld_from = from;

      r = VMessage_clone_sender(msg_fields->fld_sender, &sender);
      msg->msg_single_fields.fld_sender = sender;

      r = VMessage_clone_reply_to(msg_fields->fld_reply_to, &reply_to);
      msg->msg_single_fields.fld_reply_to = reply_to;

      r = VMessage_clone_to(msg_fields->fld_to, &to);
      msg->msg_single_fields.fld_to = to;
      
      r = VMessage_clone_cc(msg_fields->fld_cc, &cc);
      msg->msg_single_fields.fld_cc = cc;
      
      r = VMessage_clone_bcc(msg_fields->fld_bcc, &bcc);
      msg->msg_single_fields.fld_bcc = bcc;
      
      r = VMessage_clone_message_id(msg_fields->fld_message_id, &message_id);
      msg->msg_single_fields.fld_message_id = message_id;  

      r = VMessage_clone_in_reply_to(msg_fields->fld_in_reply_to, &in_reply_to);
      msg->msg_single_fields.fld_in_reply_to = in_reply_to;
      
      r = VMessage_clone_references(msg_fields->fld_references,&references);
      msg->msg_single_fields.fld_references = references;
    }
  return 0;
}

static int
VMessage_clone_orig_date(struct mailimf_orig_date *msg_orig_date, 
			 struct mailimf_orig_date **orig_date)
{
  struct mailimf_date_time *date_time;
  
  date_time = x_alloc(1, struct mailimf_date_time);

  date_time = msg_orig_date->dt_date_time;

  *orig_date = mailimf_orig_date_new(date_time);
  return 0;
}

static int
VMessage_clone_from(struct mailimf_from *msg_from,
		    struct mailimf_from **from)
{
  struct mailimf_mailbox_list *mb_list;
  struct mailimf_mailbox *mb;
  struct mailimf_mailbox *msg_mb;
  clistiter *iter;
  clist *list;

  list = clist_new();

  for (iter = clist_begin(msg_from->frm_mb_list->mb_list); iter; iter = iter->next)
    {
      msg_mb = (struct mailimf_mailbox *) iter->data;
      mb = a_Message_mailbox_clone(msg_mb);
      clist_append(list, mb);
    }
  *from = mailimf_mailbox_list_new(list);
  return 0;
}

static int
VMessage_clone_sender(struct mailimf_sender *msg_sender,
		      struct mailimf_sender **sender)
{
  struct mailimf_mailbox *mb;

  mb = a_Message_mailbox_clone(msg_sender->snd_mb);
  *sender = mailimf_sender_new(mb);
  return 0;
}

static int
VMessage_clone_reply_to(struct mailimf_reply_to *msg_reply_to,
			struct mailimf_reply_to **reply_to)
{
  struct mailimf_address_list *addr_list;

  addr_list = a_Message_address_list_clone(msg_reply_to->rt_addr_list);

  *reply_to = mailimf_reply_to_new(addr_list);

  return 0;
}

static int
VMessage_clone_to(struct mailimf_to *msg_to,
		  struct mailimf_to **to)
{
  struct mailimf_address_list *addr_list;

  addr_list = a_Message_address_list_clone(msg_to->to_addr_list);

  *to = mailimf_to_new(addr_list);

  return 0;
}

static int
VMessage_clone_cc(struct mailimf_cc *msg_cc,
		  struct mailimf_cc **cc)
{
  struct mailimf_address_list *addr_list;

  addr_list = a_Message_address_list_clone(msg_cc->cc_addr_list);

  *cc = mailimf_cc_new(addr_list);

  return 0;
}

static int
VMessage_clone_bcc(struct mailimf_bcc *msg_bcc,
		  struct mailimf_bcc **bcc)
{
  struct mailimf_address_list *addr_list;

  addr_list = a_Message_address_list_clone(msg_bcc->bcc_addr_list);

  *bcc = mailimf_bcc_new(addr_list);

  return 0;
}

static int
VMessage_clone_message_id(struct mailimf_message_id *msg_message_id,
			  struct mailimf_message_id **message_id)
{
  char *id;
  id = strdup(msg_message_id->mid_value);

  *message_id = mailimf_message_id_new(id);
  return 0;
}

static int
VMessage_clone_in_reply_to(struct mailimf_in_reply_to *msg_in_reply_to,
			   struct mailimf_in_reply_to **in_reply_to)
{
  clist *id_list;

  id_list = a_Message_msg_id_list_clone(msg_in_reply_to->mid_list);
  *in_reply_to = mailimf_in_reply_to_new(id_list);
  return 0;
}

static int
VMessage_clone_references(struct mailimf_references *msg_references,
			  struct mailimf_references **references)
{
  clist *id_list;

  id_list = a_Message_msg_id_list_clone(msg_references->mid_list);
  *references = mailimf_references_new(id_list);
  return 0;
}

XePVMessage *
a_VMessage_get_from_uid(XePFolder *folder, const char *uid)
{
  chashdatum key;
  chashdatum data;
  XePVMessage *vmessage = NULL;

  key.data = (void *) uid;
  key.len = strlen(uid);
  if (chash_get(folder->vmessages, &key, &data) == 0)
    {
      vmessage = data.data;
    }
  return vmessage;
}

void
a_VMessage_ref_folder(XePVMessage *vmessage, XePFolder *folder)
{
  char *fid;

  pthread_mutex_lock(&vmessage_mutex);

  /* FIXME: memory error check missing here */
  fid = strdup(folder->id);

  clist_append(vmessage->vfolders, folder->id);

  pthread_mutex_unlock(&vmessage_mutex);
}

int
a_VMessage_unref_folder(XePVMessage *vmessage, XePFolder *folder)
{
  clistiter *iter;

  pthread_mutex_lock(&vmessage_mutex);
  
  for (iter = clist_begin(vmessage->vfolders); iter; iter = iter->next)
    {
      if (strcmp(folder->id, (char *) iter->data) == 0)
	break;
    }

  if (iter == NULL)
    {
      pthread_mutex_unlock(&vmessage_mutex);
      return -1; /* Not found */
    }

  free(iter->data);

  clist_delete(vmessage->vfolders, iter);
  if (clist_count(vmessage->vfolders) == 0)
    a_VMessage_free(vmessage);

  pthread_mutex_unlock(&vmessage_mutex);

  return 0;
}
